from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import asyncio
from sqlalchemy.orm import Session
import joule.controllers
from tests.controllers.helpers import create_db
from joule.models import DataStream, Annotation
from joule import utilities
from joule import app_keys
from joule.constants import EndPoints, ApiErrorMessages

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestAnnotationControllerErrors(AioHTTPTestCase):

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)

        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0

        db, app[psql_key] = create_db(["/top/leaf/stream1:float32[x, y, z]",
                                     "/top/middle/leaf/stream2:int16[val1, val2]"])
        self.stream1 = db.query(DataStream).filter_by(name="stream1").one_or_none()
        self.stream2 = db.query(DataStream).filter_by(name="stream2").one_or_none()

        # add 5 event annotations to stream1
        # add 5 interval annotations to stream2
        for x in range(5):
            # start: 0, 1000, 2000, 3000, 4000
            start = utilities.timestamp_to_datetime(x * 1000)
            end = utilities.timestamp_to_datetime((x * 1000) + 200)
            a1 = Annotation(title="stream1_%d" % x, start=start)
            a2 = Annotation(title="stream2_%d" % x, start=start, end=end)
            a1.stream = self.stream1
            a2.stream = self.stream2
            db.add(a1)
            db.add(a2)
        db.commit()
        app[app_keys.db] = db
        self.db = db
        return app


    async def test_annotation_list(self):
        # errors on invalid timestamps
        resp = await self.client.request("GET", EndPoints.annotations,
                                         params=[("stream_id", self.stream1.id),
                                                 ("stream_id", self.stream2.id),
                                                 ("start", "January 1")])
        self.assertEqual(resp.status, 400)
        self.assertIn("timestamp", await resp.text())
        # must specify at least one stream_id
        resp = await self.client.request("GET", EndPoints.annotations)
        self.assertEqual(resp.status, 400)
        self.assertIn("stream_id", await resp.text())
        # non-existent streams
        bad_id = self.stream1.id + self.stream2.id
        resp = await self.client.request("GET", EndPoints.annotations,
                                         params=[("stream_id", bad_id)])
        json = await resp.json()
        self.assertEqual(0, len(json))

        bad_path = "/bad/stream/path"
        resp = await self.client.request("GET", EndPoints.annotations,
                                         params=[("stream_path", bad_path)])
        self.assertIn("does not exist", await resp.text())

        self.assertEqual(resp.status, 404)

    async def test_annotation_info(self):
        # must specify a path or an id
        resp = await self.client.request("GET", EndPoints.annotations_info)
        self.assertEqual(resp.status, 400)
        self.assertIn(ApiErrorMessages.specify_id_or_path, await resp.text())

        # stream must exist
        bad_id = self.stream1.id + self.stream2.id
        resp = await self.client.request("GET", EndPoints.annotations_info,
                                         params={"stream_id": bad_id})
        self.assertEqual(resp.status, 404)
        self.assertIn("does not exist", await resp.text())

    async def test_annotation_update(self):
        db: Session = self.app[app_keys.db]
        # must be a json request
        resp = await self.client.request("PUT", EndPoints.annotation)
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())

        # must specify id
        resp = await self.client.request("PUT", EndPoints.annotation,
                                         json={"title": "test"})
        self.assertEqual(resp.status, 400)
        self.assertIn("id", await resp.text())
        # annotation must exist
        bad_id = 1001
        resp = await self.client.request("PUT", EndPoints.annotation,
                                         json={
                                             "id": bad_id,
                                             "title": "test",
                                             "content": ""})
        self.assertEqual(resp.status, 404)
        self.assertIn("does not exist", await resp.text())

        # title cannot be empty
        old_annotation = self.db.query(Annotation).filter_by(stream_id=self.stream1.id).first()
        resp = await self.client.request("PUT", EndPoints.annotation,
                                            json={
                                                "id": old_annotation.id,
                                                "title": "",
                                                "content": "--should not be in database--"})
        self.assertEqual(resp.status, 400)
        self.assertIn("title", await resp.text())

        # make sure the annotation was not updated
        new_annotation = db.get(Annotation, old_annotation.id)
        self.assertNotEqual(new_annotation.content, "--should not be in database--")


    async def test_annotation_create(self):
        # must be a json request
        resp = await self.client.request("POST", EndPoints.annotation)
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())

        # must specify stream_id or stream_path
        resp = await self.client.request("POST", EndPoints.annotation,
                                         json={"title": "test"})
        self.assertEqual(resp.status, 400)
        self.assertIn("stream_id", await resp.text())
        # stream must exist
        bad_id = self.stream1.id + self.stream2.id
        resp = await self.client.request("POST", EndPoints.annotation,
                                         json={"stream_id": bad_id})
        self.assertEqual(resp.status, 404)
        self.assertIn("does not exist", await resp.text())
        # invalid annotation parameters
        resp = await self.client.request("POST", EndPoints.annotation,
                                         json={"stream_id": self.stream1.id,
                                               "title": "missing start",
                                               "end": 100})
        self.assertEqual(resp.status, 400)
        self.assertIn("invalid values", await resp.text())
        # NOTE: invalid id format is covered by middleware

        # title cannot be empty
        resp = await self.client.request("POST", EndPoints.annotation,
                                            json={"stream_id": self.stream1.id,
                                                "title": "",
                                                "start": 400,
                                                "end": 800,
                                                "content": ""})
        self.assertEqual(resp.status, 400)
        self.assertIn("title", await resp.text())

        # start must be before end
        resp = await self.client.request("POST", EndPoints.annotation,
                                         json={"stream_id": self.stream1.id,
                                               "title": "bad time",
                                               "start": 800,
                                               "end": 800,
                                               "content": ""})
        self.assertEqual(resp.status, 400)
        self.assertIn("start", await resp.text())

    async def test_annotation_delete(self):
        # must specify an id
        resp = await self.client.request("DELETE", EndPoints.annotation)
        self.assertEqual(resp.status, 400)
        self.assertIn("id", await resp.text())
        # id must exist
        bad_id = 1001
        resp = await self.client.request("DELETE", EndPoints.annotation, params={"id": bad_id})
        self.assertEqual(resp.status, 404)
        self.assertIn("does not exist", await resp.text())


    async def test_annotation_delete_all(self):
        # must specify a stream_id
        resp = await self.client.request("DELETE", EndPoints.stream_annotations)
        self.assertEqual(resp.status, 400)
        self.assertIn(ApiErrorMessages.specify_id_or_path, await resp.text())
        # stream_id must exist
        bad_id = self.stream1.id + self.stream2.id
        resp = await self.client.request("DELETE", EndPoints.stream_annotations, params={"stream_id": bad_id})
        self.assertEqual(resp.status, 404)
        self.assertIn("does not exist", await resp.text())
        # start and end must be timestamps
        resp = await self.client.request("DELETE", EndPoints.stream_annotations,
                                         params={"stream_id": self.stream1.id,
                                                 "start": "notatimestamp"})
        self.assertEqual(resp.status, 400)
        self.assertIn("timestamp", await resp.text())



import unittest
import sqlalchemy
import numpy as np
from joule import api, errors
from joule.utilities import datetime_to_timestamp as dt2ts

class TestDbMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()

    async def asyncTearDown(self) -> None:
        await self.node.close()

    async def test_db_connect(self):
        # should be able to read data streams from postgres
        engine = await self.node.db_connect()
        stream = await self.node.data_stream_get("/live/base")
        # compare the first row of data from Postgres to a pipe read
        with engine.connect() as conn:
            resp = conn.execute(sqlalchemy.text(f"SELECT time,elem0 from data.stream{stream.id} ORDER BY time ASC LIMIT 1"))
            row = resp.fetchone()
            sql_data = np.array([dt2ts(row[0]),row[1]])
       
        pipe = await self.node.data_read("/live/base")
        pipe_data = await pipe.read(flatten=True)
        await pipe.close()
        np.testing.assert_array_equal(sql_data, pipe_data[0])

    async def test_custom_module_schemas(self):
        # the base module requests schema basedata
        engine = await self.node.db_connect()
        import asyncio
        #while True:
        #    await asyncio.sleep(2)
        with engine.connect() as conn:
            # create a table with a name and value column in the basedata schema
            conn.execute(sqlalchemy.text("CREATE TABLE basedata.test (name TEXT, value FLOAT)"))
            # add, retrieve, and remove a row from the table
            conn.execute(sqlalchemy.text("INSERT INTO basedata.test (name,value) VALUES ('test',1.0)"))
            resp = conn.execute(sqlalchemy.text("SELECT name,value FROM basedata.test"))
            row = resp.fetchone()
            self.assertEqual(row[0], "test")
            self.assertEqual(row[1], 1.0)
            conn.execute(sqlalchemy.text("DELETE FROM basedata.test"))
            resp = conn.execute(sqlalchemy.text("SELECT COUNT(*) FROM basedata.test"))
            row = resp.fetchone()
            self.assertEqual(row[0], 0)
            conn.execute(sqlalchemy.text("DROP TABLE basedata.test"))

    async def test_db_connection_info(self):
        conn_info = await self.node.db_connection_info()
        engine = sqlalchemy.create_engine(conn_info.to_dsn())
        with engine.connect() as con:
            result = con.execute(sqlalchemy.text("SELECT 1"))
            row = result.fetchone()
            self.assertEqual(row[0], 1)

    


import asyncio
import numpy as np

from joule.models.pipes import Pipe, LocalPipe
from tests import helpers
from joule.errors import PipeError


class TestPipe(helpers.AsyncTestCase):

    def test_raises_read_errors(self):
        ''

        # input pipes must implement read
        pipe = Pipe(direction=Pipe.DIRECTION.INPUT)
        with self.assertRaises(PipeError) as e:
            asyncio.run(pipe.read())
        self.assertTrue("abstract" in "%r" % e.exception)

        # output pipes cannot be read
        pipe = Pipe(direction=Pipe.DIRECTION.OUTPUT)
        with self.assertRaises(PipeError):
            asyncio.run(pipe.read())

    def test_raises_consume_errors(self):
        ''

        # input pipes must implement consume
        pipe = Pipe(direction=Pipe.DIRECTION.INPUT)
        with self.assertRaises(PipeError) as e:
            asyncio.run(pipe.consume(100))
        self.assertTrue("abstract" in "%r" % e.exception)

        # output pipes cannot be read
        pipe = Pipe(direction=Pipe.DIRECTION.OUTPUT)
        with self.assertRaises(PipeError):
            asyncio.run(pipe.consume(100))

    def test_raises_write_errors(self):
        ''

        # input pipes cannot write
        pipe = Pipe(direction=Pipe.DIRECTION.INPUT)
        with self.assertRaises(PipeError):
            asyncio.run(pipe.write(np.array([1, 2, 3, 4])))

        # output pipes must implement write
        pipe = Pipe(direction=Pipe.DIRECTION.OUTPUT)
        with self.assertRaises(PipeError) as e:
            asyncio.run(pipe.write(np.array([1, 2, 3, 4])))
        self.assertTrue("abstract" in "%r" % e.exception)

    def test_raises_cache_errors(self):
        ''
        # input pipes cannot cache
        pipe = Pipe(direction=Pipe.DIRECTION.INPUT)
        with self.assertRaises(PipeError):
            pipe.enable_cache(100)
        with self.assertRaises(PipeError):
            asyncio.run(pipe.flush_cache())

        # output pipes must implement caching
        pipe = Pipe(direction=Pipe.DIRECTION.OUTPUT)
        with self.assertRaises(PipeError) as e:
            pipe.enable_cache(100)
        self.assertTrue("abstract" in "%r" % e.exception)
        with self.assertRaises(PipeError):
            asyncio.run(pipe.flush_cache())
        self.assertTrue("abstract" in "%r" % e.exception)

    def test_raises_close_interval_errors(self):
        ''

        # input pipes cannot close intervals
        pipe = Pipe(direction=Pipe.DIRECTION.INPUT)
        with self.assertRaises(PipeError):
            asyncio.run(pipe.close_interval())

        # output pipes must implement close_interval
        pipe = Pipe(direction=Pipe.DIRECTION.OUTPUT)
        with self.assertRaises(PipeError) as e:
            asyncio.run(pipe.close_interval())
        self.assertTrue("abstract" in "%r" % e.exception)

    def test_checks_dtype(self):
        for layout in ['invalid', 'bad_3', 'float_abc']:
            pipe = Pipe(layout=layout)
            with self.assertRaises(ValueError) as e:
                _ = pipe.dtype
            self.assertTrue("layout" in "%r" % e.exception)

    def test_end_of_interval_default(self):
        # end of interval flag is false by default
        pipe = Pipe(direction=Pipe.DIRECTION.INPUT)
        self.assertFalse(pipe.end_of_interval)

    def test_raises_dtype_errors(self):
        pipe = Pipe(layout="int16_4")
        # data for a different stream type
        data1 = helpers.create_data("float32_3")
        # invalid array structure (must be 2D)
        data2 = np.ones((3, 3, 3))
        # invalid structured array
        data3 = np.array([('timestamp', 1, 2, 3), ('bad', 1, 2, 3)])
        for data in [data1, data2, data3]:
            with self.assertRaises(PipeError):
                _ = pipe._apply_dtype(data)

    def test_subscribe(self):
        LAYOUT = "int16_4"
        # cannot subscribe to input pipes
        input_pipe = Pipe(layout=LAYOUT, direction=Pipe.DIRECTION.INPUT)
        output_pipe = Pipe(layout=LAYOUT, direction=Pipe.DIRECTION.OUTPUT)
        subscriber = Pipe(layout=LAYOUT, direction=Pipe.DIRECTION.OUTPUT)
        with self.assertRaises(PipeError):
            input_pipe.subscribe(subscriber)
        # can subscribe to output pipes
        unsubscribe = output_pipe.subscribe(subscriber)
        self.assertIn(subscriber, output_pipe.subscribers)
        # can unsubscribe
        unsubscribe()
        self.assertNotIn(subscriber, output_pipe.subscribers)

    def test_read_all_flatten(self):
        asyncio.run(self._test_read_all_flatten())

    async def _test_read_all_flatten(self):
        # This test is a bit tricky because it needs to run in an unbroken
        # event loop since it creates LocalPipes outside of the test
        LAYOUT = "int32_3"
        LENGTH = 1000
        ''

        # raises exception if the pipe is empty
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        my_pipe.close_nowait()
        with self.assertRaises(PipeError):
            await my_pipe.read_all(flatten=True)
        # read_all empties pipe and closes it, regardless of intervals
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        my_pipe.close_nowait()
        actual_data = await my_pipe.read_all(flatten=True)
        expected_data = np.hstack((test_data1, test_data2))
        expected_data = np.c_[expected_data['timestamp'][:, None], expected_data['data']]
        np.testing.assert_array_equal(actual_data, expected_data)
        self.assertTrue(my_pipe.closed)

        # read_all only add maxrows to the pipe
        # (less than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        actual_data = await my_pipe.read_all(maxrows=103, flatten=True)
        expected_data = test_data1[:103]
        expected_data = np.c_[expected_data['timestamp'][:, None], expected_data['data']]
        np.testing.assert_array_equal(actual_data, expected_data)
        self.assertTrue(my_pipe.closed)
        # (more than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        actual_data = await my_pipe.read_all(maxrows=LENGTH + 101, flatten=True)
        expected_data = np.hstack((test_data1, test_data2[:101]))
        expected_data = np.c_[expected_data['timestamp'][:, None], expected_data['data']]
        np.testing.assert_array_equal(expected_data, actual_data)
        self.assertTrue(my_pipe.closed)

        # read_all raises an exception if the pipe has more than maxrows
        # (less than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        with self.assertRaises(PipeError):
            await my_pipe.read_all(maxrows=LENGTH + 101,
                                                     error_on_overflow=True, flatten=True)
        self.assertTrue(my_pipe.closed)
        # (more than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        my_pipe.close_nowait()
        with self.assertRaises(PipeError):
            await my_pipe.read_all(maxrows=LENGTH + 101,
                                                     error_on_overflow=True, flatten=True)
        self.assertTrue(my_pipe.closed)

    def test_read_all_no_flatten(self):
        asyncio.run(self._test_read_all_no_flatten())

    async def _test_read_all_no_flatten(self):
        LAYOUT = "int32_3"
        LENGTH = 1000
        ''

        # raises exception if the pipe is empty
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        my_pipe.close_nowait()
        with self.assertRaises(PipeError):
            await my_pipe.read_all()

        # read_all empties pipe and closes it, regardless of intervals
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        my_pipe.close_nowait()
        actual_data = await my_pipe.read_all()
        expected_data = np.hstack((test_data1, test_data2))
        np.testing.assert_array_equal(actual_data, expected_data)
        self.assertTrue(my_pipe.closed)

        # read_all only add maxrows to the pipe
        # (less than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        actual_data = await my_pipe.read_all(maxrows=103)
        expected_data = test_data1[:103]
        np.testing.assert_array_equal(actual_data, expected_data)
        self.assertTrue(my_pipe.closed)
        # (more than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        actual_data = await my_pipe.read_all(maxrows=LENGTH + 101)
        expected_data = np.hstack((test_data1, test_data2[:101]))
        np.testing.assert_array_equal(expected_data, actual_data)
        self.assertTrue(my_pipe.closed)

        # read_all raises an exception if the pipe has more than maxrows
        # (less than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        with self.assertRaises(PipeError):
            await my_pipe.read_all(maxrows=LENGTH + 101, error_on_overflow=True)
        self.assertTrue(my_pipe.closed)
        # (more than one read)
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)
        my_pipe.close_nowait()
        with self.assertRaises(PipeError):
            await my_pipe.read_all(maxrows=LENGTH + 101, error_on_overflow=True)
        self.assertTrue(my_pipe.closed)



import unittest
from joule import api, errors
from constants import *

class TestBackupIngestCopy(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()
        """
        └── original
            ├── stream1
            └── stream2
            └── events
        """
        
    async def asyncTearDown(self):
        await self.node.close()

    async def test_backup_ingest(self):
        self.assertTrue(True)
        print("here!")
        print(t1)


import warnings
from aiohttp.test_utils import unused_port
import dateparser
from click.testing import CliRunner
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

warnings.simplefilter('always')


class TestDelete(FakeJouleTestCase):

    def test_data_delete(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        start_str = "20 January 2015 12:00"
        end_str = "1 hour ago"
        result = runner.invoke(main, ['data', 'delete',
                                      '/folder/src',
                                      '--start', start_str,
                                      '--end', end_str])
        self.assertEqual(result.exit_code, 0)
        (path, start, end) = self.msgs.get()
        self.assertEqual(path, '/folder/src')
        self.assertEqual(int(start), int(dateparser.parse(start_str).timestamp() * 1e6))
        # the relative time argument should grow
        self.assertLess(int(end), int(dateparser.parse(end_str).timestamp() * 1e6))
        self.stop_server()

    def test_invalid_parameters(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        # test invalid start and end times
        result = runner.invoke(main, ['data', 'delete', '/folder/src', '--start', 'bad_time'])
        self.assertIn("Error", result.output)
        result = runner.invoke(main, ['data', 'delete', '/folder/src', '--end', 'bad_time'])
        self.assertIn("Error", result.output)
        # must specify time range or all
        result = runner.invoke(main, ['data', 'delete', '/folder/src'])
        self.assertIn("Error", result.output)
        # cannot specify both time range and all
        result = runner.invoke(main, ['data', 'delete', '/folder/src',
                                       '--start', 'now', '--all'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_stream_does_not_exist(self):
        server = FakeJoule()
        server.response = "stream does not exist"
        server.http_code = 404
        server.stub_data_delete = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'delete', '/folder/src'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        server.stub_data_delete = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'delete', '/folder/src', '--all'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



from click.testing import CliRunner
import os
import warnings
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

VERSION_JSON = os.path.join(os.path.dirname(__file__), 'version.json')
DBINFO_JSON = os.path.join(os.path.dirname(__file__), 'dbinfo.json')

warnings.simplefilter('always')


class TestInfo(FakeJouleTestCase):

    def test_joule_info(self):
        server = FakeJoule()
        with open(VERSION_JSON, 'r') as f:
            server.response = f.read()
        with open(DBINFO_JSON, 'r') as f:
            server.dbinfo_response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['node', 'info'])
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure the version is displayed
        self.assertIn("0.8.2", output)
        self.stop_server()


    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['node', 'info'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



import asyncio
import numpy as np
import random
import logging
import asyncpg
import socket
import psutil

from joule.models import DataStream, pipes
from joule.errors import EmptyPipeError
from joule.models.data_store import psql_helpers
from joule.models.data_store.errors import DataError
import joule.utilities

Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')
MAX_CHUNK_INTERVAL = 1000 * 1000 * 60 * 60 * 24 * 365 # 1 year in microseconds


class Inserter:

    def __init__(self, pool: asyncpg.pool.Pool, stream: DataStream, insert_period: float,
                 cleanup_period: float, merge_gap: int=0):
        self.pool = pool
        self.stream = stream
        self._data_rate_buffer = []
        # TODO: make this configurable, right now it is computed automatically
        self._chunk_interval = 0  # duration in seconds, 0 means not set
        # round cleanup_period to a multiple of insert_period
        if insert_period == 0 or cleanup_period < insert_period:
            self.cleanup_interval = 1  # clean with every insert
        else:
            self.cleanup_interval = np.ceil(cleanup_period / insert_period)
        # add offsets to the period to distribute traffic
        self.insert_period = insert_period + insert_period * random.random() * 0.5
        self.decimator: Decimator = None
        # merge this insertion with the previous data if the timestamps are <= merge_gap us appart
        self.merge_gap = merge_gap

    async def run(self, pipe: pipes.Pipe) -> None:
        """insert stream data from the queue until the queue is empty"""

        # lazy stream creation
        try:
            async with self.pool.acquire() as conn:
                await psql_helpers.create_stream_table(conn, self.stream)
        except asyncio.CancelledError:
            return

        # close the beginning of the data insert
        first_insert = True
        # track the last timestamp inserted
        last_ts = None
        ticks = 0
        try:
            while True:
                await asyncio.sleep(self.insert_period)
                data = await pipe.read()
                await self._measure_data_rate(data)
                try:
                    async with self.pool.acquire() as conn:
                        # there might be an interval break and no new data
                        if len(data) > 0:
                            if first_insert:
                                first_insert = False
                                await psql_helpers.close_interval(conn, self.stream, data['timestamp'][0] - 1)
                                # if the difference between the current data and this new data
                                # is less than merge_gap apart remove the interval boundary
                                if self.merge_gap > 0:
                                    await psql_helpers.remove_interval_breaks(conn, self.stream, 
                                                                              start=data['timestamp'][0]-self.merge_gap,
                                                                              end=data['timestamp'][0])
                                #closest_ts = await psql_helpers.get_closest_ts(conn, self.stream, data['timestamp'][0])
                                #if closest_ts is not None and data['timestamp'][0] - closest_ts > self.merge_gap:
                                #    await psql_helpers.close_interval(conn, self.stream, data['timestamp'][0] - 1)
                                #    print("closing interval with previous data")
                                #elif closest_ts is None:
                                #    print("no earlier ts in data")
                                #else:
                                #    print(f"merging new data with pervious interval: {data['timestamp'][0] - closest_ts} <= {self.merge_gap}")
                                #    print(f"TODO: remove any previous intervals")
                            if not joule.utilities.misc.timestamps_are_monotonic(data, last_ts, self.stream.name):
                                raise DataError("Non-monotonic timestamps in new data")
                            if not joule.utilities.misc.validate_values(data):
                                raise DataError("Invalid values (NaN) in new data")
                            last_ts = data['timestamp'][-1]
                            # lazy initialization of decimator
                            if self.stream.decimate and self.decimator is None:
                                self.decimator = Decimator(self.stream, 1, 4, self._chunk_interval)
                            psql_bytes = psql_helpers.data_to_bytes(data)
                            try:
                                await conn.copy_to_table("stream%d" % self.stream.id,
                                                         schema_name='data',
                                                         format='binary',
                                                         source=psql_bytes)
                                #print("-----BEGIN-------")
                                #print(f"wrote {data['timestamp']} timestamps to stream{self.stream.id}")
                                #display the byteio object as a hex string
                                #print(psql_helpers.data_to_bytes(data).getvalue().hex())
                                #print("-------END-----")
                            except asyncpg.exceptions.UniqueViolationError as e:
                                raise DataError(e)

                            # this was successful so consume the data
                            pipe.consume(len(data))
                            # decimate the data
                            if self.decimator is not None:
                                await self.decimator.process(conn, data)
                        # check for interval breaks
                        if pipe.end_of_interval and last_ts is not None:
                            await psql_helpers.close_interval(conn, self.stream, last_ts)
                            if self.decimator is not None:
                                self.decimator.close_interval()
                        ticks += 1
                        if ticks % self.cleanup_interval == 0:
                            await self.cleanup(conn)
                except (asyncpg.exceptions.PostgresConnectionError, socket.error) as e:
                    log.error(f"Timescale inserter: [{str(e)}, trying again in 2 seconds")
                    await asyncio.sleep(2)

        except (EmptyPipeError, asyncio.CancelledError):
            pass

    async def cleanup(self, conn: asyncpg.Connection):
        if self.stream.keep_us == DataStream.KEEP_ALL:
            return
        tables = await psql_helpers.get_table_names(conn, self.stream, with_schema=False)
        # ts is milliseconds UNIX timestamp
        keep_s = self.stream.keep_us // 1e6
        for table in tables:
            if 'interval' in table:
                # drop all boundaries before the cutoff
                cutoff = joule.utilities.timestamp_to_datetime(
                    joule.utilities.time_now() - self.stream.keep_us)
                query = "DELETE FROM data.%s WHERE time < '%s'" % (table, cutoff)
            else:
                query = "SELECT drop_chunks('data.%s',older_than => interval '%d seconds')" % (table, keep_s)
            await conn.execute(query)

    async def update_chunk_interval(self, chunk_interval: int) -> None:
        self._chunk_interval = min(chunk_interval, MAX_CHUNK_INTERVAL)
        async with self.pool.acquire() as conn:
            await psql_helpers.update_chunk_interval(conn, f"data.stream{self.stream.id}",
                                                     self._chunk_interval)
            if self.decimator is not None:
                await self.decimator.update_chunk_interval(conn, self._chunk_interval)

    async def _measure_data_rate(self, data: np.ndarray) -> None:
        if self._chunk_interval != 0:
            return  # already measured
        self._data_rate_buffer += (data['timestamp']).tolist()
        if len(self._data_rate_buffer) < 200:
            return  # not enough data
        # determine the data rate by using the median of the differences
        # between timestamps
        diffs = np.diff(self._data_rate_buffer)
        data_rate = 1e6 / np.median(diffs) * data.dtype.itemsize
        total_memory = psutil.virtual_memory().total
        # set chunk size to 5% of available memory, this allows ~5 active streams
        self._chunk_interval = (int(total_memory * 0.05) / data_rate)*1e6 # in microseconds
        hours = self._chunk_interval / (1e6*60*60)
        memory = int(total_memory*0.05)/1e6
        #print(f"setting target chunk interval for {self.stream.name} [{self.stream.id}] to {hours}  ~ {memory}MB")
        #print(f"Data rate: {data_rate} bytes/sec, chunk interval: {self._chunk_interval} ms")
        await self.update_chunk_interval(self._chunk_interval)


class Decimator:

    def __init__(self, stream: DataStream, from_level: int, factor: int,
                 chunk_interval: int, debug=False):
        self.stream = stream
        self.level = from_level * factor
        self.table_name = "stream%d_%d" % (stream.id, self.level)
        self.full_table_name = "data." + self.table_name
        if from_level > 1:
            self.again = True
        else:
            self.again = False
        self.factor = factor
        self.layout = stream.decimated_layout
        self.buffer = []
        self.path_created = False
        self.chunk_interval = min(chunk_interval, MAX_CHUNK_INTERVAL)
        self.child: Decimator = None
        self.debug = debug
        if self.debug:
            print("creating decim level %d" % self.level)
        # hold off to rate limit traffic
        self.holdoff = 0  # random.random()

    async def update_chunk_interval(self, conn: asyncpg.Connection, chunk_interval):
        self.chunk_interval = min(chunk_interval, MAX_CHUNK_INTERVAL)
        await psql_helpers.update_chunk_interval(conn, self.full_table_name, self.chunk_interval)
        if self.child is not None:
            await self.child.update_chunk_interval(conn, self.chunk_interval * 4)

    async def process(self, conn: asyncpg.Connection, data: np.ndarray) -> None:
        """decimate data and insert it, retry on error"""
        if not self.path_created:
            await psql_helpers.create_decimation_table(conn, self.stream, self.level)
            if self.chunk_interval > 0:
                await psql_helpers.update_chunk_interval(conn, self.full_table_name, self.chunk_interval)
            self.path_created = True

        decim_data = self._process(data)
        if self.debug:
            print("\t level %d: %d rows" % (self.level, len(decim_data)))
        if len(decim_data) == 0:
            return

        # lazy initialization of child
        if self.child is None:
            self.child = Decimator(self.stream, self.level, self.factor,
                                   self.chunk_interval * 4, self.debug)

        psql_bytes = psql_helpers.data_to_bytes(decim_data)
        await conn.copy_to_table(self.table_name,
                                 schema_name='data',
                                 format='binary',
                                 source=psql_bytes)
        await self.child.process(conn, decim_data)

    def close_interval(self):
        self.buffer = []
        if self.child is not None:
            self.child.close_interval()

    def _process(self, sarray: np.ndarray) -> np.ndarray:

        # flatten structured array
        data = np.c_[sarray['timestamp'][:, None], sarray['data']]

        # check if there is old data
        if len(self.buffer) != 0:
            # append the new data onto the old data
            data = np.concatenate((self.buffer, data))
        (n, m) = data.shape

        # Figure out which columns to use as the input for mean, min, and max,
        # depending on whether this is the first decimation or we're decimating
        # again.  Note that we include the timestamp in the means.
        if self.again:
            c = (m - 1) // 3
            # e.g. c = 3
            # ts mean1 mean2 mean3 min1 min2 min3 max1 max2 max3
            mean_col = slice(0, c + 1)
            min_col = slice(c + 1, 2 * c + 1)
            max_col = slice(2 * c + 1, 3 * c + 1)
        else:
            mean_col = slice(0, m)
            min_col = slice(1, m)
            max_col = slice(1, m)

        # Discard extra rows that aren't a multiple of factor
        n = n // self.factor * self.factor

        if n == 0:  # not enough data to work with, save it for later
            self.buffer = data
            return np.array([])

        trunc_data = data[:n, :]
        # keep the leftover data
        self.buffer = np.copy(data[n:, :])

        # Reshape it into 3D so we can process 'factor' rows at a time
        trunc_data = trunc_data.reshape(n // self.factor, self.factor, m)

        # Fill the result
        out = np.c_[np.mean(trunc_data[:, :, mean_col], axis=1),
                    np.min(trunc_data[:, :, min_col], axis=1),
                    np.max(trunc_data[:, :, max_col], axis=1)]

        # structure the array
        width = np.shape(out)[1] - 1
        dtype = np.dtype([('timestamp', '<i8'), ('data', '<f4', width)])
        sout = np.zeros(out.shape[0], dtype=dtype)
        sout['timestamp'] = out[:, 0]
        sout['data'] = out[:, 1:]
        # insert the data into the database
        return sout



import unittest
from datetime import datetime, timezone
from joule.models import (EventStream, Folder, event_stream)
from joule.models.data_store.event_store import StreamInfo


class TestStream(unittest.TestCase):
    def setUp(self):
        pass

    def test_attributes(self):
        my_folder = Folder(name="test")
        my_stream = EventStream(name='test', description='a test',
                                event_fields={'field1': 'string', 'field2': 'numeric'},
                                keep_us=100,
                                chunk_duration_us=0,
                                updated_at=datetime.now(timezone.utc))
        my_stream.folder = my_folder
        new_attrs = {
            "name": "new_n",
            "description": "new_d",
            "event_fields": {"test_val": "string"},
            "keep_us": 1000,
            "chunk_duration_us": 100
        }
        last_update = my_stream.updated_at
        my_stream.update_attributes(new_attrs)
        # make sure all attributes are updated
        self.assertEqual(my_stream.name, new_attrs["name"])
        self.assertEqual(my_stream.description, new_attrs["description"])
        self.assertEqual(my_stream.chunk_duration_us, new_attrs["chunk_duration_us"])
        self.assertEqual(my_stream.event_fields, new_attrs["event_fields"])
        # the updated at should be more recent
        self.assertGreater(my_stream.updated_at, last_update)

        # touching also increments the updated_at
        last_update = my_stream.updated_at
        my_stream.touch()
        self.assertGreater(my_stream.updated_at, last_update)
        self.assertEqual(my_stream.updated_at, my_folder.updated_at)

    def test_representations(self):
        my_stream = EventStream(name='test', description='a test',
                                event_fields={'field1': 'string', 'field2': 'numeric'},
                                keep_us=100,
                                chunk_duration_us=0,
                                updated_at=datetime.now(timezone.utc))
        json_resp = my_stream.to_json()
        self.assertEqual(json_resp["name"], my_stream.name)
        self.assertEqual(json_resp["description"], my_stream.description)
        self.assertEqual(json_resp["updated_at"], my_stream.updated_at.isoformat())
        self.assertEqual(json_resp["keep_us"], my_stream.keep_us)
        self.assertEqual(json_resp["chunk_duration_us"], my_stream.chunk_duration_us)
        self.assertDictEqual(json_resp['event_fields'], my_stream.event_fields)

        # has a string representation
        self.assertIn(my_stream.name, str(my_stream))

        # can be recovered from JSON
        new_stream = event_stream.from_json(json_resp)
        self.assertEqual(new_stream.name, my_stream.name)
        self.assertEqual(new_stream.description, my_stream.description)
        # updated_at is not serialized, instead it is set to now
        self.assertGreater(new_stream.updated_at, my_stream.updated_at)
        self.assertEqual(new_stream.keep_us, my_stream.keep_us)
        self.assertEqual(new_stream.chunk_duration_us, my_stream.chunk_duration_us)
        self.assertDictEqual(new_stream.event_fields, my_stream.event_fields)

    def test_handles_stream_info(self):
        my_stream = EventStream(name='test', description='a test',
                                event_fields={'field1': 'string', 'field2': 'numeric'},
                                keep_us=100,
                                chunk_duration_us=0,
                                updated_at=datetime.now(timezone.utc))
        my_stream.id = 100
        stream_info = StreamInfo(start=0, end=100, event_count=10, total_time=100, bytes=100)
        json_resp = my_stream.to_json(info={my_stream.id: stream_info})
        self.assertDictEqual(json_resp['data_info'], stream_info.to_json())

        # ignored when info is not available
        json_resp = my_stream.to_json(info={101: stream_info})
        self.assertNotIn('data_info', json_resp)




import click
from joule.cli.lazy_group import LazyGroup

@click.group(name="node",
             cls=LazyGroup,
             lazy_subcommands={
                 "info": "joule.cli.node.info.node_info",
                 "list": "joule.cli.node.list.node_list",
                 "add": "joule.cli.node.add.node_add",
                 "delete": "joule.cli.node.delete.node_delete",
                 "default": "joule.cli.node.default.node_default"
             })
def node():
    """Configure local settings."""
    pass  




from datetime import datetime, timezone
import numpy as np
import os
import pdb
import configparser
import asyncio
import unittest
import testing.postgresql
import psycopg2
from sqlalchemy import create_engine, pool
from sqlalchemy.orm import Session
from collections import deque
from typing import Optional
from icecream import ic
from joule.models import DataStream, Element, Base, Pipe
from joule.models.pipes import LocalPipe
from joule.errors import EmptyPipeError, PipeError


def create_data(layout: str,
                length=100,
                step=1000,  # in us
                start=1476152086000000):  # 10 Oct 2016 10:15PM
    """Create a random block of data with [layout] structure"""
    ts = np.arange(start, start + step * length, step, dtype=np.uint64)

    # Convert to structured array
    (ltype, lcount, dtype) = parse_layout(layout)
    sarray = np.zeros(len(ts), dtype=dtype)
    if "float" in ltype:
        data = np.random.rand(length, lcount)
    elif "uint" in ltype:
        data = np.random.randint(0, high=100, size=(
            length, lcount), dtype=dtype[1].base)
    else:
        data = np.random.randint(-50, high=50,
                                 size=(length, lcount), dtype=dtype[1].base)

    sarray['timestamp'] = ts
    # Need the squeeze in case sarray['data'] is 1 dimensional
    sarray['data'] = data
    return sarray


def create_stream(name, layout, id=0) -> DataStream:
    (ltype, lcount, dtype) = parse_layout(layout)
    datatype = DataStream.DATATYPE[ltype.upper()]

    return DataStream(name=name, datatype=datatype, id=id,
                      elements=[Element(name="e%d" % j, index=j,
                                        display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(lcount)],
                      updated_at=datetime.now(timezone.utc))


def to_chunks(data, chunk_size):
    """Yield successive MAX_BLOCK_SIZE chunks of data."""
    for i in range(0, len(data), chunk_size):
        yield data[i:i + chunk_size]


def parse_layout(layout):
    ltype = layout.split('_')[0]
    lcount = int(layout.split('_')[1])
    if ltype.startswith('int'):
        atype = '<i' + str(int(ltype[3:]) // 8)
    elif ltype.startswith('uint'):
        atype = '<u' + str(int(ltype[4:]) // 8)
    elif ltype.startswith('float'):
        atype = '<f' + str(int(ltype[5:]) // 8)
    else:
        raise ValueError("bad layout")
    dtype = np.dtype([('timestamp', '<i8'), ('data', atype, (lcount,))])
    return ltype, lcount, dtype


def parse_configs(config_str):
    config = configparser.ConfigParser()
    config.read_string(config_str)
    return config


def mock_stream_info(streams):
    """pass in array of stream_info's:
       [['/test/path','float32_3'],
       ['/test/path2','float32_5'],...]
       returns a function to mock stream_info as a side_effect"""

    def stream_info(path):
        for stream in streams:
            if stream[0] == path:
                return [stream]
        return []

    return stream_info


class AsyncTestCase(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        self.loop.set_debug(True)
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        closed = self.loop.is_closed()
        if not closed:
            self.loop.call_soon(self.loop.stop)
            self.loop.run_forever()
            self.loop.close()
        asyncio.set_event_loop(None)

        # close any remaining pipes, uvloop tends to leave
        # pipes after the loop is closed
        for str_fd in set(os.listdir('/proc/self/fd/')):
            fd = int(str_fd)
            try:
                fstat = os.fstat(fd)
                # if stat.S_ISFIFO(fstat.st_mode):
                #    os.close(fd)
            except OSError:
                pass

        import psutil
        proc = psutil.Process()
        # print("[%d] fds" % proc.num_fds())


Postgresql = testing.postgresql.PostgresqlFactory(cache_initialized_db=True)


class DbTestCase(unittest.TestCase):

    def setUp(self):
        self.postgresql = Postgresql()  # testing.postgresql.Postgresql()
        db_url = self.postgresql.url()
        conn = psycopg2.connect(db_url)
        cur = conn.cursor()
        cur.execute("CREATE SCHEMA metadata")
        conn.commit()
        conn.close()
        engine = create_engine(db_url, poolclass=pool.NullPool)
        Base.metadata.create_all(engine)
        self.db = Session(bind=engine)

    def tearDown(self):
        self.db.close()
        self.postgresql.stop()


class TestingPipe(Pipe):
    # reads come out in the same blocks as the writes
    # ...wraps LocalPipe

    def __init__(self, layout: str, name: str = None, stream=None):
        super().__init__(name=name, layout=layout, stream=stream)

        self._pipe = LocalPipe(layout, name, stream)
        self._closed = False
        self.data_blocks: deque = deque()

    async def write(self, data):
        if self._closed:
            raise PipeError("Cannot write to a closed pipe")
        self.data_blocks.append(data)

    def write_nowait(self, data):
        if self._closed:
            raise PipeError("Cannot write to a closed pipe")
        self.data_blocks.append(data)

    def close_interval_no_wait(self):
        self.data_blocks.append(None)

    async def close_interval(self):
        self.data_blocks.append(None)

    def consume(self, rows):
        return self._pipe.consume(rows)

    def reread_last(self):
        self._pipe.reread_last()

    @property
    def end_of_interval(self):
        return self._pipe.end_of_interval

    async def read(self, flatten=False):
        # first write a block to the pipe if there are any waiting
        if len(self.data_blocks) > 0:
            block = self.data_blocks.popleft()
            if block is None:
                await self._pipe.close_interval()
            else:
                await self._pipe.write(block)
        elif self._closed:
            await self._pipe.close()
        # now return the result of the inner pipe's read
        return await self._pipe.read(flatten)

    def is_empty(self):
        return self._pipe.is_empty()

    async def close(self):
        self._closed = True


class TestingPipeOld(Pipe):
    # reads come out in the same blocks as the writes

    def __init__(self, layout: str, name: str = None, stream=None):
        super().__init__(name=name, layout=layout, stream=stream)
        self.data_blocks: deque = deque()
        self.last_block: Optional[np.ndarray] = None
        self._last_read = False  # flag to indicate pipe only has previously read data (see input_pipe.py)
        self.interval_break = False
        self._reread = False

    async def write(self, data):
        self.write(data)
        
    def write_nowait(self, data):
        sarray = self._apply_dtype(data)
        self.data_blocks.append(sarray)

    def close_interval_no_wait(self):
        self.data_blocks.append(None)

    async def close_interval(self):
        self.data_blocks.append(None)

    def consume(self, rows):
        if self.last_block is None:
            raise Exception("Nothing to consume")
        if rows > len(self.last_block):
            raise Exception("Cannot consume %d rows, only %d available" % (
                rows, len(self.last_block)))
        if rows == len(self.last_block):
            self.last_block = None
        else:
            self.last_block = self.last_block[rows:]

    def reread_last(self):
        self._reread = True

    @property
    def end_of_interval(self):
        return self.interval_break

    async def read(self, flatten=False):
        if flatten:
            raise Exception("Not Implemented")
        if self._reread:
            self._reread = False
            if self.last_block is None or len(self.last_block) == 0:
                raise PipeError("No data left to reread")
            return self.last_block

        if len(self.data_blocks) == 0 and self.last_block is None:
            raise EmptyPipeError()
        if len(self.data_blocks) != 0:
            block = self.data_blocks.popleft()
            if len(self.data_blocks) == 0:
                self.interval_break = True
                self._last_read = True
            elif self.data_blocks[0] is None:
                self.data_blocks.popleft()
                self.interval_break = True
            else:
                self.interval_break = False

            if self.last_block is not None:
                self.last_block = np.hstack((self.last_block, block))
            else:
                self.last_block = block
        return self.last_block

    def is_empty(self):
        if self._reread:
            return False
        if self._last_read:
            return True
        if len(self.data_blocks) == 0 and self.last_block is None:
            return True
        return False



import unittest
import asyncio
import numpy as np

from joule.models.pipes import LocalPipe
from tests import helpers
from joule.errors import EmptyPipeError


class TestLocalPipeBasicOperation(unittest.TestCase):
    def test_normal_operation(self):
        # running through the data, each element is returned once
        # every call to read returns some data or an EmptyPipe error
        async def _run():
            LAYOUT = "float32_3"
            LENGTH = 100
            pipe = LocalPipe(layout="float32_3")
            tx_data = helpers.create_data(LAYOUT, length=LENGTH)
            # write a block and close off the pipe
            await pipe.write(tx_data)
            await pipe.close()
            # read it back
            rx_data = await pipe.read()
            np.testing.assert_array_equal(tx_data, rx_data)
            # don't consume it so you can read it agai
            pipe.consume(len(rx_data[:500]))
            # read it back again
            rx_data = await pipe.read()
            np.testing.assert_array_equal(tx_data[500:], rx_data)
            # the pipe should be empty now
            self.assertTrue(pipe.is_empty())
            with self.assertRaises(EmptyPipeError):
                await pipe.read()



from tests.api import mock_session
import unittest
from joule.api.node import TcpNode


from joule.constants import EndPoints


class TestFollowerApi(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        # no URL or event loop
        self.node = TcpNode('mock_node', 'http://url', 'api_key')
        self.session = mock_session.MockSession()
        self.node.session = self.session

    async def test_follower_delete_by_name(self):
        await self.node.follower_delete("follower")
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.follower)
        self.assertEqual(self.session.request_data, {'name': "follower"})

    async def test_follower_delete_by_object(self):       
        # can delete by node object
        follower = TcpNode('follower', 'http://url', 'api_key')
        await self.node.follower_delete(follower)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.follower)
        self.assertEqual(self.session.request_data, {'name': "follower"})

    ## follower list is covered by CLI tests


import unittest
import testing.postgresql
import asyncpg
import random
import tempfile
import shutil
import os
import sys
import asyncio

from joule.models.data_store.event_store import EventStore, StreamInfo
from joule.models.event_stream import EventStream
from joule.utilities.time import human_to_timestamp as h2ts
from joule.utilities.time import timestamp_to_human as ts2h

SQL_DIR = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'src', 'joule', 'sql')


class TestEventStore(unittest.IsolatedAsyncioTestCase):
    use_default_loop = False
    forbid_get_event_loop = True

    async def asyncSetUp(self):
        # set up the psql database
        self.psql_dir = tempfile.TemporaryDirectory()
        # run this asynchronously to avoid blocking the event loop
        def start_postgresql_instance():
            return testing.postgresql.Postgresql(base_dir=self.psql_dir.name)
        self.postgresql = await asyncio.to_thread(start_postgresql_instance)
        await asyncio.to_thread(self.postgresql.stop)

        # now that the directory structure is created, customize the *.conf file
        src = os.path.join(os.path.dirname(__file__), "postgresql.conf")
        dest = os.path.join(self.psql_dir.name, "data", "postgresql.conf")
        shutil.copyfile(src, dest)
        # restart the database
        self.postgresql = await asyncio.to_thread(start_postgresql_instance)

        self.db_url = self.postgresql.url()
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        await conn.execute("CREATE EXTENSION IF NOT EXISTS timescaledb CASCADE ")
        await conn.execute("CREATE USER joule_module")
        # load custom functions
        for file in os.listdir(SQL_DIR):
            file_path = os.path.join(SQL_DIR, file)
            with open(file_path, 'r') as f:
                await conn.execute(f.read())
        await conn.close()

    async def asyncTearDown(self):
        await asyncio.to_thread(self.postgresql.stop)
        self.psql_dir.cleanup()

    async def test_runner(self):
        tests = [self._test_histogram,
                 self._test_basic_upsert_extract_remove,
                 self._test_time_bound_queries,
                 self._test_remove_bound_queries,
                 self._test_json_queries,
                 self._test_limit_queries,
                 self._test_info,
                 self._test_count
                 ]
        for test in tests:
            conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
            await conn.execute("DROP SCHEMA IF EXISTS data CASCADE")
            await conn.execute("CREATE SCHEMA data")
            await conn.execute("GRANT ALL ON SCHEMA data TO public")

            self.store = EventStore(self.db_url)
            await self.store.initialize()
            await conn.close()
            await test()
            # simulate the nose2 test output
            sys.stdout.write('o')
            await self.store.close()
            sys.stdout.flush()

    async def _test_info(self):
        streams = []
        for i in range(20):
            event_stream = EventStream(id=i, name=f'test_stream{i}')
            await self.store.create(event_stream)
            streams.append(event_stream)
            if i % 2 == 0:  # only half the streams have events
                events = [{'start_time': j * 100,
                           'end_time': j * 100 + 10,
                           'content': {'title': "Event%d" % j,
                                       'stream': event_stream.name}
                           } for j in range(i * 1000, i * 1000 + 20 + i)]
            else:
                events = []
            await self.store.upsert(event_stream, events)

        def _expected_info(stream_ids):
            results = {}
            for x in stream_ids:
                if x % 2 != 0:
                    results[x] = StreamInfo(None, None, 0, 0, 0)
                else:
                    start = x * 1000 * 100
                    end = (x * 1000 + 20 + x - 1) * 100 + 10
                    results[x] = StreamInfo(start, end, 20 + x, end - start, 0)
            return results

        # info from one stream
        stream_info = await self.store.info([streams[4]])
        self.assertDictEqual(stream_info, _expected_info([4]))
        # info about multiple streams
        stream_info = await self.store.info(streams[:2])
        self.assertDictEqual(stream_info, _expected_info([0, 1]))
        # info about a lot of streams (trigger bulk select)
        stream_info = await self.store.info(streams[1:-1])
        ids = [s.id for s in streams[1:-1]]
        self.assertDictEqual(stream_info, _expected_info(ids))

    async def _test_count(self):
        event_stream = EventStream(id=1, name='test_stream')
        await self.store.create(event_stream)
        # create 20 events, 5 with a tag of A and 15 with a tag of B
        events = [{'start_time': i * 100,
                   'end_time': i * 100 + 10,
                   'content': {'title': "Event%d" % i,
                               'stream': event_stream.name,
                               'tag': 'A' if(i < 5) else 'B'}
                   } for i in range(20)]
        await self.store.upsert(event_stream, events)
        # basic count operation
        count = await self.store.count(event_stream)
        self.assertEqual(count, 20)
        # count operation with filter to only return 'A' events
        count = await self.store.count(event_stream, json_filter=[[['tag', 'is', 'A']]])
        self.assertEqual(count, 5)
        # count operation with time bounds to only return events between 1005 and  < 1500
        count = await self.store.count(event_stream, start=1005, end=1500, include_on_going_events=False)
        self.assertEqual(count, 4)
        # includes on going events gets the event at 1000
        count = await self.store.count(event_stream, start=1005, end=1500, include_on_going_events=True)
        self.assertEqual(count, 5)


    async def _test_histogram(self):
        def _count_events(items):
            acc = 0
            for item in items:
                acc += item['content']['count']
            return acc

        # generate 200 events with start times between Aug 1 and Aug 2 2021
        #[h2ts("2021-08-01T00:00:00"), h2ts("2021-08-02T00:00:00")]
        # use microsecond timestamp to improve execution speed (avoid asyncio warning)
        bounds1 = [1627790400000000, 1627876800000000]
        times = [random.randint(bounds1[0], bounds1[1]) for _ in range(200)]
        event_stream = EventStream(id=1, name='test_stream')
        await self.store.create(event_stream)
        events1 = [{'start_time': ts,
                    'end_time': ts + random.randint(int(10e6), int(100e6)),
                    'content': {'block': 1}} for ts in times]
        # generate 150 events with start times between Aug 3 and Aug 4 2021
        #[h2ts("2021-08-03T00:00:00"), h2ts("2021-08-04T00:00:00")] ^-- see above note
        bounds2 = [1627963200000000, 1628049600000000]
        times = [random.randint(bounds2[0], bounds2[1]) for _ in range(150)]
        events2 = [{'start_time': ts,
                    'end_time': ts + random.randint(int(10e6), int(100e6)),
                    'content': {'block': 2}} for ts in times]
        await self.store.upsert(event_stream, events1 + events2)
        # now get the histogram for the whole range
        hist = await self.store.histogram(event_stream)
        self.assertEqual(_count_events(hist), len(events1 + events2))
        # get the histogram for block 1 by time
        hist = await self.store.histogram(event_stream, end=bounds1[1], nbuckets=50)
        self.assertEqual(_count_events(hist), len(events1))
        # expect the right number of buckets
        self.assertEqual(len(hist), 50)
        # get the histogram for block 2 by time
        hist = await self.store.histogram(event_stream, start=bounds2[0], nbuckets=150)
        self.assertEqual(_count_events(hist), len(events2))
        # expect the right number of buckets
        self.assertEqual(len(hist), 150)
        # get the histogram for block 2 by time, but specify both bounds
        hist = await self.store.histogram(event_stream, start=bounds2[0], end=bounds2[1], nbuckets=150)
        self.assertEqual(_count_events(hist), len(events2))
        # expect the right number of buckets
        self.assertEqual(len(hist), 150)
        # get the histogram for block 1 by JSON filter
        hist = await self.store.histogram(event_stream, json_filter=[[['block', 'eq', 1]]])
        self.assertEqual(_count_events(hist), len(events1))
        # expect the right number of buckets
        self.assertEqual(len(hist), 100)
        # retrieving the earlier block and the later JSON filter should return no records
        hist = await self.store.histogram(event_stream, json_filter=[[['block', 'eq', 2]]],
                                          end=bounds1[1])
        self.assertEqual(_count_events(hist), 0)
        # expect the right number of buckets
        self.assertEqual(len(hist), 100)

    async def _test_basic_upsert_extract_remove(self):
        event_stream1 = EventStream(id=1, name='test_stream1')
        await self.store.create(event_stream1)
        events1 = [{'start_time': i * 100,
                    'end_time': i * 100 + 10,
                    'content': {'title': "Event%d" % i,
                                'stream': event_stream1.name}
                    } for i in range(20)]
        await self.store.upsert(event_stream1, events1)
        event_stream2 = EventStream(id=2, name='test_stream2')
        await self.store.create(event_stream2)
        events2 = [{'start_time': (i * 100) + 3,
                    'end_time': (i * 100) + 13,
                    'content': {'title': "Event%d" % i,
                                'stream': event_stream2.name}
                    } for i in range(20)]
        await self.store.upsert(event_stream2, events2)
        # now try to read them back
        rx_events = await self.store.extract(event_stream1)
        # they should match but now have an id field
        ids = set([e["id"] for e in rx_events])
        self.assertEqual(len(ids), len(events1))
        # update the first event
        rx_events[0]['content']['title'] = 'updated'
        rx2_events = await self.store.upsert(event_stream1, rx_events)
        rx3_events = await self.store.extract(event_stream1)
        self.assertListEqual(rx_events, rx2_events)
        self.assertListEqual(rx2_events, rx3_events)
        # ...the two event streams are independent
        rx_events = await self.store.extract(event_stream2)
        # they should match but now have an id field (which is not included in the equality test)
        # the received events will also have an event_stream_id field
        # so we put this field into events2 for comparison
        for e in events2:
            e['event_stream_id']=event_stream2.id
        self.assertListEqual(events2, rx_events)
        # now remove them
        await self.store.remove(event_stream1)
        rx_events = await self.store.extract(event_stream1)
        self.assertListEqual(rx_events, [])
        # ...the two event streams are independent
        rx_events = await self.store.extract(event_stream2)
        self.assertListEqual(events2, rx_events)

    async def _test_time_bound_queries(self):
        event_stream1 = EventStream(id=1, name='test_stream1')
        await self.store.create(event_stream1)
        early_events = [{'start_time': i * 100,
                         'end_time': i * 100 + 10,
                         'event_stream_id': event_stream1.id,
                         'content': {'title': "Event%d" % i,
                                     'stream': event_stream1.name}
                         } for i in range(20)]
        mid_events = [{'start_time': i * 100,
                       'end_time': i * 100 + 10,
                       'event_stream_id': event_stream1.id,
                       'content': {'title': "Event%d" % i,
                                   'stream': event_stream1.name}
                       } for i in range(30, 40)]
        late_events = [{'start_time': i * 100,
                        'end_time': i * 100 + 10,
                        'event_stream_id': event_stream1.id,
                        'content': {'title': "Event%d" % i,
                                    'stream': event_stream1.name}
                        } for i in range(50, 60)]
        await self.store.upsert(event_stream1, early_events + mid_events + late_events)

        # now try to read them back
        rx_events = await self.store.extract(event_stream1, start=3000, end=4500)
        self.assertListEqual(rx_events, mid_events)
        rx_events = await self.store.extract(event_stream1, end=4500)
        self.assertListEqual(rx_events, early_events + mid_events)
        rx_events = await self.store.extract(event_stream1, start=3000)
        self.assertListEqual(rx_events, mid_events + late_events)

        # now try offset timebounds to verify include_ongoing_events
        rx_events = await self.store.extract(event_stream1, start=3010, end=4500, include_on_going_events=False)
        self.assertListEqual(rx_events, mid_events[1:]) # skips first event
        rx_events = await self.store.extract(event_stream1, start=3010, end=4500, include_on_going_events=True)
        self.assertListEqual(rx_events, mid_events) # includes first event


        # now try limited queries
        rx_events = await self.store.extract(event_stream1, start=3000, limit=10)
        self.assertListEqual(rx_events, mid_events[:10])
        rx_events = await self.store.extract(event_stream1, end=4500, limit=10)
        self.assertListEqual(rx_events, mid_events[-10:])

    async def _test_remove_bound_queries(self):
        event_stream1 = EventStream(id=20, name='test_stream1')
        await self.store.create(event_stream1)

        early_events = [{'start_time': i * 100,
                         'end_time': i * 100 + 10,
                         'event_stream_id': event_stream1.id,
                         'content': {'title': "Event%d" % i,
                                     'stream': event_stream1.name}
                         } for i in range(20)]
        mid_events = [{'start_time': i * 100,
                       'end_time': i * 100 + 10,
                       'event_stream_id': event_stream1.id,
                       'content': {'title': "Event%d" % i,
                                   'stream': event_stream1.name}
                       } for i in range(30, 40)]
        late_events = [{'start_time': i * 100,
                        'end_time': i * 100 + 10,
                        'event_stream_id': event_stream1.id,
                        'content': {'title': "Event%d" % i,
                                    'stream': event_stream1.name}
                        } for i in range(50, 60)]
        await self.store.upsert(event_stream1, early_events + mid_events + late_events)

        # remove the middle
        await self.store.remove(event_stream1, start=3000, end=4500)
        rx_events = await self.store.extract(event_stream1)
        self.assertListEqual(rx_events, early_events + late_events)
        # remove the beginning
        await self.store.remove(event_stream1, end=3000)
        rx_events = await self.store.extract(event_stream1)
        self.assertListEqual(rx_events, late_events)
        # remove the end
        await self.store.remove(event_stream1, start=4500)
        rx_events = await self.store.extract(event_stream1)
        self.assertListEqual(rx_events, [])

    async def _test_json_queries(self):
        event_stream1 = EventStream(id=20, name='test_stream1')
        await self.store.create(event_stream1)

        early_events = [{'start_time': i * 100,
                         'end_time': i * 100 + 10,
                         'event_stream_id': event_stream1.id,
                         'content': {'title': "Event%d" % i,
                                     'type': 'early',
                                     'stream': event_stream1.name}
                         } for i in range(20)]
        mid_events = [{'start_time': i * 100,
                       'end_time': i * 100 + 10,
                       'event_stream_id': event_stream1.id,
                       'content': {'title': "Event%d" % i,
                                   'type': 'middle',
                                   'stream': event_stream1.name}
                       } for i in range(30, 40)]
        late_events = [{'start_time': i * 100,
                        'end_time': i * 100 + 10,
                        'event_stream_id': event_stream1.id,
                        'content': {'title': "Event%d" % i,
                                    'stream': event_stream1.name}
                        } for i in range(50, 60)]
        await self.store.upsert(event_stream1, early_events + mid_events + late_events)

        middle_events = await self.store.extract(event_stream1, json_filter=[[['type', 'is', 'middle']]])
        self.assertListEqual(mid_events, middle_events)
        specific_event = await self.store.extract(event_stream1, json_filter=[[['type', 'is', 'middle'],
                                                                               ['title', 'is', 'Event30']]])
        self.assertEqual(mid_events[0], specific_event[0])

    async def _test_limit_queries(self):
        event_stream1 = EventStream(id=20, name='test_stream1')
        await self.store.create(event_stream1)

        events = [{'start_time': i * 100,
                   'end_time': i * 100 + 10,
                   'event_stream_id': event_stream1.id,
                   'content': {'title': "Event%d" % i,
                               'type': 'early',
                               'stream': event_stream1.name}
                   } for i in range(20)]
        await self.store.upsert(event_stream1, events)

        first_events = await self.store.extract(event_stream1, limit=5)
        self.assertListEqual(events[:5], first_events)

# from https://stackoverflow.com/questions/5844672/delete-an-element-from-a-dictionary
# def _remove_id(events):
#    return [{i: e[i] for i in e if i != 'id'} for e in events]



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp
import uuid

import joule.controllers
from joule.constants import EndPoints
from tests.controllers.helpers import MockStore
from joule import app_keys
from joule.utilities import ConnectionInfo

class TestRootController(AioHTTPTestCase):

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        app[app_keys.data_store] = MockStore()
        app[app_keys.name] = "test_suite"
        app[app_keys.uuid] = uuid.uuid4()
        self.connection_info = ConnectionInfo(username="test",
                                            password="test", 
                                            port=5432, 
                                            database="test", 
                                            host="test")
        app[app_keys.module_connection_info] = self.connection_info
        return app


    async def test_index(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", "/")
        msg = await resp.text()
        self.assertEqual(resp.status, 200)
        # 'joule' is somewhere in the response string
        self.assertTrue('joule' in msg.lower())


    async def test_version(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.version)
        version = await resp.text()
        self.assertEqual(resp.status, 200)
        # version is some non-empty string
        self.assertTrue(len(version) > 0)
        # check JSON version
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.version_json)
        data = await resp.json()
        self.assertEqual(resp.status, 200)
        # version is some non-empty string
        self.assertTrue(len(data['version']) > 0)


    async def test_dbinfo(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.db_info)
        dbinfo = await resp.json()
        self.assertEqual(resp.status, 200)
        # check one of the keys to make sure this is a dbinfo object
        self.assertTrue('path' in dbinfo)

    async def test_db_connection(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.db_connection)
        dbinfo = await resp.json()
        self.assertEqual(resp.status, 200)
        self.assertDictEqual(dbinfo, self.connection_info.to_json())



from click.testing import CliRunner
import os
import warnings
import json
import copy
import asyncio
import re
from unittest import SkipTest

from joule.models import DataStream
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

EVENT_INFO = os.path.join(os.path.dirname(__file__), 'event_info.json')
warnings.simplefilter('always')


class TestEventInfo(FakeJouleTestCase):

    def test_show_info(self):
        server = FakeJoule()
        with open(EVENT_INFO, 'r') as f:
            server.response = f.read()
        server.stub_event_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['event', 'info', '-e', '/folder_1/random'])
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly

        self.stop_server()
